#pragma once

#include <format>
#include <optional>
#include <print>
#include <stdexcept>
#include <string_view>
#include <unordered_map>
#include <vector>

class ArgParser {
public:
  void add(std::string_view shortOpt, std::string_view longOpt, const std::string_view desc, const bool hasValue = true, const bool isRequired = false,
           std::string_view defaultValue = "") {
    const Argument arg{shortOpt, longOpt, desc, defaultValue, hasValue, isRequired};
    m_arguments.emplace_back(arg);

    m_opts.emplace(shortOpt, std::ref(m_arguments.back()));
    m_opts.emplace(longOpt, std::ref(m_arguments.back()));
  }

  [[nodiscard]] bool parse(const int argc, const char* argv[]) const {
    if (argc == 1) {
      printHelp();
      return false;
    }

    for (int i = 1; i < argc; ++i) {
      std::string_view arg = argv[i];

      if (arg == "-h" || arg == "--help") {
        printHelp();
        return false;
      }

      if (auto argument = findArgument(arg); argument.has_value()) {
        auto& argRef     = argument->get();
        argRef.isPresent = true;

        if (argRef.requiresValue) {
          if (i + 1 >= argc) {
            throw std::invalid_argument(std::format("Missing value for: {}", arg));
          }
          argRef.value = argv[++i];
        }
      } else {
        throw std::invalid_argument(std::format("Unknown option: {}", arg));
      }
    }
    for (const auto& arg : m_opts) {
      const auto& argRef = arg.second.get();
      if (argRef.isRequired && !argRef.isPresent) {
        throw std::invalid_argument(std::format("Required argument: [{}, {}] not provided", argRef.shortOpt, argRef.longOpt));
      }
    }
    return true;
  }

  template <typename T>
  T get(const std::string_view arg) const {
    if (const auto argument = findArgument(arg); argument.has_value()) {
      return convertToType<T>(argument->get().value.value());
    } else {
      throw std::runtime_error(std::format("{} has no value", arg));
    }
  }

  bool has(std::string_view arg) const {
    if (const auto argument = findArgument(arg); argument.has_value()) {
      return argument->get().isPresent;
    }
    return false;
  }

private:
  struct Argument {
    std::string_view                shortOpt;
    std::string_view                longOpt;
    std::string_view                description;
    std::optional<std::string_view> value;
    bool                            requiresValue;
    bool                            isRequired = false;
    bool                            isPresent  = false;
  };

  template <typename T>
  T convertToType(const std::string_view value) const {
    if constexpr (std::is_same_v<T, bool>) {
      if (value == "true")
        return true;
      if (value == "false")
        return false;
      throw std::runtime_error("Invalid boolean string");
    } else if constexpr (std::is_arithmetic_v<T>) {
      T result{};
      auto [ptr, ec] = std::from_chars(value.data(), value.data() + value.size(), result);
      if (ec != std::errc()) {
        throw std::runtime_error(std::format("Failed to convert the value: {}", value));
      }
      return result;
    } else if constexpr (std::is_same_v<T, std::string>) {
      return std::string(value);
    } else if constexpr (std::is_same_v<T, std::string_view>) {
      return value;
    } else {
      throw std::runtime_error("Unsupported type for conversion");
    }
  }

  std::optional<std::reference_wrapper<Argument>> findArgument(const std::string_view opt) const {
    if (const auto it = m_opts.find(opt); it != m_opts.end()) {
      return it->second;
    }
    return std::nullopt;
  }

  void printHelp() const {
    std::println("Usage: [options]");
    for (const auto& arg : m_arguments) {
      std::println("   {}, {}: {}", arg.shortOpt, arg.longOpt, arg.description);
    }
  }

  std::vector<Argument>                                                  m_arguments;
  std::unordered_map<std::string_view, std::reference_wrapper<Argument>> m_opts;
};
